package simulation;

/**
 * Model a molecular participant in reactions as a name
 * and a number of 'instances'.
 * A Molecule is an entirely passive repositories of a count
 * and whether it is a boundary species.
 * 
 * @author David J. Barnes (d.j.barnes@kent.ac.uk)
 *
 */
public class Molecule
{
	// The name of the molecule.
	private String name;
	// The number of instances in the system.
	private int number;
	// Whether this is a boundary species.
	private boolean isBoundary;
	
	/**
	 * Create a molecule.
	 * @param name The name of the molecule.
	 * @param number The number of instances.
	 * @param isBoundary Whether it is a boundary species.
	 */
	public Molecule(String name, int number, boolean isBoundary)
	{
		this.name = name;
		this.number = number;
		this.isBoundary = isBoundary;
	}
	
	/**
	 * Create a molecule which is not a boundary species.
	 * @param name The name of the molecule.
	 * @param number The number of instances.
	 */
	public Molecule(String name, int number)
	{
		this(name, number, false);
	}
	
	/**
	 * Create a molecule with no instances and which is not a boundary species.
	 * @param name The name of the molecule.
	 */
	public Molecule(String name)
	{
		this(name, 0, false);
	}
	
	/**
	 * Return the molecule's name.
	 * @return The molecule's name.
	 */
	public String getName()
	{
		return name;
	}
	
	/**
	 * Return the molecular count.
	 * @return The number of instances.
	 */
	public int getNumber()
	{
		return number;
	}
	
	/**
	 * If this is not a boundary species, adjust the amount of this
	 * particular molecule, plus or minus.
	 * @param amount The delta amount, plus or minus.
	 */
	public void delta(int amount)
	{
		if(!isBoundary) {
			number += amount;
			assert number >= 0;
		}
	}
	
	/**
	 * Return whether this is a boundary species or not.
	 * @return true if a boundary species, false otherwise.
	 */
	public boolean isBoundarySpecies()
	{
		return isBoundary;
	}
	
	/**
	 * Return details of this molecule.
	 * @return Details of this molecule.
	 */
	public String toString()
	{
		return name + (isBoundary? "*" : "") + " (" + number + ")";
	}
	
}
