package simulation;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Random;
import java.util.Set;
import java.util.HashSet;
import java.util.List;

/**
 * Model a reaction as a combination of reactants yielding products.
 * Either reactants or products may be empty, but not both.
 * 
 * @author David J. Barnes (d.j.barnes@kent.ac.uk)
 *
 */
public class Reaction implements Comparable<Reaction>
{
	// Shared random number generator.
	private static Random rand = new Random(); 
	// The reaction's reactants. May be empty.
	private List<Reactant> reactants;
	// The reaction's products. May be empty.
	private List<Product> products;
	// The reaction's dependency set.
	private Set<Molecule> dependsOn;
	// The reaction's affects set.
	private Set<Molecule> affects;
	// Reaction constant.
	// For generality, it should be possible for this to vary.
	// This can be achieved by overriding the calculatePropensity method.
	private double reactionConstant;
	// propensity = combinations * reactionConstant;
	private double propensity;
	// The next time of this reaction
	private double nextTime;
	
	// Provision for the propensity becoming zero.
	// The last non-zero propensity.
	private double lastNonZeroPropensity;
	// The time at which the propensity became zero.
	private double zeroPropensityTime;
	
	// The reaction's index in the time-ordered queue.
	private int queueIndex;
	
	/**
	 * Set up for the reactants and products of a reaction.
	 * The reaction is not complete until the reactants and/or products
	 * have been added.
	 * @param reactionConstant The reaction constant
	 */
	public Reaction(double reactionConstant)
	{
		this.reactants = new ArrayList<>();
		this.products = new ArrayList<>();
		this.dependsOn = new HashSet<>();
		this.affects = new HashSet<>();
		this.reactionConstant = reactionConstant;
		this.propensity = 0;
		this.zeroPropensityTime = 0;
		this.queueIndex = 0;	// Not known on creation.
	}
	
	/**
	 * Add a reactant with multiplicity 1.
	 * @param m The reactant.
	 */
	public void addReactant(Molecule m)
	{
		addReactant(m, 1);
	}
	
	/**
	 * Add a reactant with given multiplicity.
	 * @param m The reactant.
	 * @param multiplicity Positive multiplicity.
	 */
	public void addReactant(Molecule m, int multiplicity)
	{
		assert multiplicity > 0;
		reactants.add(new Reactant(m, multiplicity));
		if(!m.isBoundarySpecies()) {
			addToDependsOn(m);
			addToAffects(m);
		}
	}
	
	/**
	 * Add a product with multiplicity 1.
	 * @param m The product.
	 */
	public void addProduct(Molecule m)
	{
		addProduct(m, 1);
	}
	
	/**
	 * Add a product with given multiplicity.
	 * @param m The product.
	 * @param multiplicity Positive multiplicity.
	 */
	public void addProduct(Molecule m, int multiplicity)
	{
		assert multiplicity > 0;
		products.add(new Product(m, multiplicity));
		addToAffects(m);
	}
	
	/**
	 * Adjust the next time by the given ratio of oldPropensity over propensity.
	 * @param time The current time
	 * @param oldPropensity The old propensity
	 */
	public void adjustTime(double time, double oldPropensity)
	{
		assert(propensity != 0);
		// Take account of a possible previous zero propensity.
		if(oldPropensity != 0) {
			nextTime = time + (nextTime - time) * oldPropensity / propensity;
		}
		else {
			// The reaction has moved from zero propensity to non-zero.
			if(lastNonZeroPropensity > 0) {
				nextTime = time + (nextTime - zeroPropensityTime) * (lastNonZeroPropensity / propensity);
			}
			else {
				// There hasn't been a previous non-zero propensity.
				calculateNextTime(time);
			}
		}
	}
	
	/**
	 * Calculate the next time for this reaction.
	 * @param time The current time.
	 */
	public void calculateNextTime(double time)
	{
		if(propensity != 0) {
			double deltaT = (1.0 / propensity * Math.log(1.0 / rand.nextDouble()));
			nextTime = time + deltaT;	
		}
		else {
			// Impossible reaction.
			nextTime = Double.POSITIVE_INFINITY;
			// No valid previous last non-zero propensity.
			lastNonZeroPropensity = 0;
		}
	}
	
	/**
	 * Return the propensity.
	 * @return The propensity.
	 */
	public double getPropensity()
	{
		return propensity;
	}
	
	/**
	 * Calculate and set the value of the propensity.
	 * Note if the propensity becomes zero.
	 * @param time The current time.
	 * @return The new value of the propensity.
	 */
	public double calculatePropensity(double time)
	{
		double oldPropensity = propensity;
		propensity = propensityCalculation();
		
		if(propensity == 0 && oldPropensity != 0) {
			// Remember the last non-zero value and when the change happened.
			lastNonZeroPropensity = oldPropensity;
			zeroPropensityTime = time;
		}
		return propensity;
	}
	
	/**
	 * Calculate the propensity from the molecular combinations
	 * and reaction constant.
	 * @return The calculated propensity.
	 */
	protected double propensityCalculation()
	{
		return calculateCombinations() * getReactionConstant();
	}
	
	/**
	 * Return the reaction constant for this reaction.
	 * @return The reaction constant for this reaction.
	 */
	public double getReactionConstant()
	{
		return reactionConstant;
	}

	/**
	 * Return the putative next time of this reaction.
	 * @return The putative next time of this reaction.
	 *         May be INFINITY.
	 */
	public double getNextTime()
	{
		if(propensity != 0) {
			return nextTime;
		}
		else {
			// It cannot currently occur.
			return Double.POSITIVE_INFINITY;
		}
	}
	

	/**
	 * Return the set of molecules this reaction depends upon.
	 * @return The set of molecules this reaction depends upon.
	 */
	public Set<Molecule> getDependsOn() 
	{
		return dependsOn;
	}

	/**
	 * Return the set of molecules this reaction affects.
	 * @return The set of molecules this reaction affects.
	 */
	public Set<Molecule> getAffects() 
	{
		return affects;
	}

	/**
	 * Apply this reaction, as long as there are sufficient reactants.
	 */
	public void react()
	{
		// Check that there are sufficient reactants.
		for(Reactant r : reactants) {
			if(r.getMultiplicity() > r.getMolecule().getNumber()) {
				throw new RuntimeException("Impossible reaction selected: " + this);
			}
		}
		
		// Consume the reactants.
		for(Reactant r : reactants) {
			r.consume();
		}
		// Produce the products.
		for(Product p : products) {
			p.produce();
		}
	}
	
	/**
	 * Return a string version of this reaction.
	 * @return A string version of this reaction.
	 */
	public String toString()
	{
		StringBuilder builder = new StringBuilder();
		Iterator<Reactant> itR = reactants.iterator();
		if(itR.hasNext()) {
			builder.append(itR.next());
			while(itR.hasNext()) {
				builder.append(" + ");
				builder.append(itR.next());
			}
		}
		builder.append(" -> ");
		builder.append("(" + getPropensity()+") ");
		Iterator<Product> itP = products.iterator();
		if(itP.hasNext()) {
			builder.append(itP.next());
			while(itP.hasNext()) {
				builder.append(" + ");
				builder.append(itP.next());
			}
		}
		return builder.toString();
	}
	
	/**
	 * Add the molecule to the list of molecules this reaction depends on.
	 * @param m The molecule this reaction depends on.
	 */
	protected void addToDependsOn(Molecule m)
	{
		dependsOn.add(m);
	}
	
	/**
	 * Calculate the number of combinations based on the numbers of participating and
	 * available reactants.
	 * @return The number of combinations.
	 */
	private double calculateCombinations()
	{
		// If there are no explicit reactants then 1.0 is returned.
		double combinations = 1.0;
		for(Reactant r : reactants) {
			int m = r.getMultiplicity();
			if(m > 1) {
				// Avoid factorial calculations.
				int numerator = 1;
				int denominator = 1;
				for(int times = 0; times < m; times++) {
					numerator *= r.getMolecule().getNumber() - times;
					denominator *= m - times;
				}
				h *= numerator / denominator;
			}
			else {
				combinations *= r.getMolecule().getNumber();
			}
		}
		return combinations;
	}
	
	/**
	 * Add the molecule to the list of molecules affected by this reaction.
	 * However, if m is catalytic in this reaction then it should not occur
	 * in affects.
	 * @param m The molecule affected by this reaction.
	 */
	private void addToAffects(Molecule m)
	{
		// Look for catalytic reactions.
		boolean catalytic = false;
		Reactant reactant = null;
		Product product = null;
		for(Reactant r : reactants) {
			if(r.getMolecule() == m) {
				reactant = r;
			}
		}
		for(Product p : products) {
			if(p.getMolecule() == m) {
				product = p;
			}
		}
		if(reactant != null && product != null) {
			if(reactant.getMultiplicity() == product.getMultiplicity() &&
					reactant.isConsumed()) {
				catalytic = true;
			}
		}

		if(!catalytic){
			affects.add(m);
		}
		else {
			// Remove any previous addition.
			affects.remove(m);
		}
	}

	/**
	 * @return -1, 0 or 1 depending on whether the reaction time is less than, equal to
	 *         or greater than that of the argument.
	 */
	public int compareTo(Reaction r)
	{
		double diff = nextTime - r.nextTime;
		return diff < 0 ? -1 : diff > 0 ? 1 : 0;
	}

	/** 
	 * Set the index in the time-ordered queue for this reaction.
	 * @param queueIndex The index in the time-ordered queue.
	 */
	public void setQueueIndex(int queueIndex)
	{
		this.queueIndex = queueIndex;
	}

	/**
	 * Return the index in the time-ordered queue for this reaction.
	 * @return The index in the time-ordered queue for this reaction.
	 */
	public int getQueueIndex()
	{
		return queueIndex;
	}
	
}
