package simulation;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Maintain a set of interacting reactions.
 *
 * @author David J. Barnes (d.j.barnes@kent.ac.uk)
 */
public class ReactionSet 
{
	// The reactions.
	private List<Reaction> reactions;
	// A convenience map of molecule names.
	private Map<String, Molecule> molecules;

	// The dependency graph between reactions.
	private Map<Reaction, Set<Reaction>> dependencyGraph;
	// The reactions ordered according to time of next reaction.
	// For a node at element index, parentIndex = (index - 1) / 2;
	// leftChildIndex = 2 * index + 1; rightChildIndex = leftChildIndex + 1
	private Reaction[] timeOrderedQueue;
	
	/**
	 * Create a reaction set ready to be initialized
	 * via initReactionSet.
	 */
	public ReactionSet()
	{
		this.reactions = new ArrayList<>();
		this.molecules = new HashMap<>();
		
		// Create the following once the number of reactions is known.
		this.dependencyGraph = null;
		this.timeOrderedQueue = null;
	}
		
	/**
	 * Initialize the required reaction set.
	 */
	public void init()
	{
		buildDependencyGraph();
		
		for(Reaction r : reactions) {
			r.calculatePropensity(0);
			r.calculateNextTime(0);
		}
		
		initQueue();
	}
	
	/**
	 * Clear the reaction set from any previous run.
	 */
	public void clear()
	{
		this.reactions.clear();
		this.molecules.clear();
	}
	
	/**
	 * Find the reaction with the earliest next time. 
	 * @return The selected reaction.
	 */
	public Reaction findNextReaction()
	{
		return timeOrderedQueue[0];
	}
	
	/**
	 * Update the system according to the effect of the last
	 * reaction.
	 * @param lastReaction The last reaction.
	 * @param time The time of the last reaction.
	 */
	public void update(Reaction lastReaction, double time)
	{
		for(Reaction r : dependencyGraph.get(lastReaction)) {
			double oldA = r.getPropensity();
			double newA = r.calculatePropensity(time);
			
			if(r == lastReaction) {
				r.calculateNextTime(time);
			}
			else {
				// Adjust the putative time of an affected reaction.
				if(newA != 0) {
					r.adjustTime(time, oldA);
				}
			}
			// Update r within the queue.
			updateQueue(r);
		}
	}
	
	/**
	 * Add a molecule to the reaction set.
	 * @param m The molecule to be added.
	 */
	public void addMolecule(Molecule m)
	{
		molecules.put(m.getName(), m);		
	}
	
	/**
	 * Get the molecule with the given name.
	 * @param name The name of the molecule
	 * @return Return the given molecule.
	 */
	public Molecule getMolecule(String name)
	{
		return molecules.get(name);
	}
	
	/**
	 * Add a new Reaction to the set of reactions.
	 * @param The reaction to be added.
	 * @return The reaction added.
	 */
	public Reaction addReaction(Reaction r)
	{
		reactions.add(r);
		return r;
	}
	
	/**
	 * Build the dependency graph for the reactions.
	 * NB: This does not take account of indirect dependencies
	 *     that might arise through species-dependent reaction 'constants'.
	 */
	private void buildDependencyGraph() 
	{
		this.dependencyGraph = new HashMap<>(reactions.size());
		// Map of molecules to the list of reactions depending on them.
		Map<Molecule, List<Reaction>> moleculeDependencyList = 
			new HashMap<>(molecules.size());
		
		// Build a map of which reactions depend on which molecules.
		for(Reaction target : reactions) {
			Set<Molecule> dependsOn = target.getDependsOn();
			for(Molecule m : dependsOn) {
				List<Reaction> dependency = moleculeDependencyList.get(m);
				if(dependency == null) {
					dependency = new ArrayList<>();
					moleculeDependencyList.put(m, dependency);
				}
				dependency.add(target);
			}
		}

		// For each reaction, identify which molecules it affects
		// and which reactions those molecules depend upon.
		for(Reaction source : reactions) {
			// Although the dependencies should have set semantics, the
			// building process will avoid duplicates, so we can get away
			// with a list rather than a set.
			Set<Reaction> affectedReactions = new HashSet<>();
			dependencyGraph.put(source, affectedReactions);
			
			// Add the self dependency.
			affectedReactions.add(source);
			
			Set<Molecule> affects = source.getAffects();
			// Identify the reactions affected.
			for(Molecule m : affects) {
				List<Reaction> dependency = moleculeDependencyList.get(m);
				if(dependency != null) {
					// This reaction affects all the dependent reactions.
					affectedReactions.addAll(dependency);
				}
			}
		}		
	}
		
	/**
	 * Show the dependency graph for the reactions.
	 */
	private void showDependencyGraph()
	{
		System.out.println("Dependencies");
		for(Reaction source : reactions) {
			System.out.println(source);
			Set<Reaction> affectedReactions = dependencyGraph.get(source);
			for(Reaction target : affectedReactions) {
				System.out.println("    " + target);
			}			
		}		
		System.out.println();
	}

	/**
	 * Initialize the priority queue for the reactions.
	 */
	private void initQueue()
	{
		// Create the structure for the priority queue.
		int numberOfReactions = reactions.size();
		this.timeOrderedQueue = new Reaction[numberOfReactions];
		
		for(int index = 0; index < numberOfReactions; index++) {
			insertIntoQueue(index, reactions.get(index));
		}
	}
	
	/**
	 * Update the queue after a reaction to reflect the fact that the
	 * given reaction has an updated time, which might not be
	 * the earliest.
	 * @param r The reaction whose position is to be checked.
	 */
	private void updateQueue(Reaction r)
	{
		int index = r.getQueueIndex();
		int parentIndex = (index - 1) / 2;
		Reaction parent = timeOrderedQueue[parentIndex];
		if(r.getNextTime() < parent.getNextTime()) {
			swap(parentIndex, index, r);
			updateQueue(r);
		}			
		else {
			// Compare against any children.
			int leftChildIndex = 2 * index + 1;
			if(leftChildIndex < timeOrderedQueue.length) {
				Reaction leftChild = timeOrderedQueue[leftChildIndex];
				// Will need to consider the right child (if any) as well.
				Reaction rightChild = null;
				int rightChildIndex = leftChildIndex + 1;
				if(rightChildIndex < timeOrderedQueue.length) {
					rightChild = timeOrderedQueue[rightChildIndex];
				}
				
				if(r.getNextTime() <= leftChild.getNextTime()) {
					// Don't swap with the left.
					if(rightChild != null) {
						// Try the right.
						if(r.getNextTime() > rightChild.getNextTime()) {
							// Swap with the right child.
							swap(index, rightChildIndex, rightChild);
							updateQueue(r);
						}
						else {
							// No swap needed.
						}
					}
					else {
						// No right child.
					}
				}
				else {
					// Should be swapped with one of the children.
					// See which has min(nextTime).
					if(rightChild != null) {
						// Try the right.
						if(rightChild.getNextTime() < leftChild.getNextTime()) {
							// Swap with the right.
							swap(index, rightChildIndex, rightChild);
							updateQueue(r);
						}
						else {
							// Swap with the left.
							swap(index, leftChildIndex, leftChild);
							updateQueue(r);
						}
					}
					else {
						// No right child. Swap with the left.
						swap(index, leftChildIndex, leftChild);
						updateQueue(r);
					}						
				}
			}				
		}
	}
	
	/**
	 * Insert reaction into the queue at the given index.
	 * Ensure that the queue is maintained in priority order.
	 * @param index The insertion point in the queue.
	 * @param reaction The reaction to be inserted.
	 */
	private void insertIntoQueue(int index, Reaction reaction)
	{
		timeOrderedQueue[index] = reaction;
		reaction.setQueueIndex(index);
		int parentIndex = (index - 1) / 2;
		Reaction parent = timeOrderedQueue[parentIndex];
		while(index > 0 && parent.getNextTime() > reaction.getNextTime()) {
			// Swap.
			swap(parentIndex, index, reaction);			
			index = parentIndex;
			parentIndex = (index - 1) / 2;
			parent = timeOrderedQueue[parentIndex];
		}
	}
	
	/**
	 * Swap a parent and child in the queue.
	 * @param parentIndex The index of the parent.
	 * @param childIndex The index of the child.
	 * @param child The child to be moved.
	 */
	private void swap(int parentIndex, int childIndex, Reaction child)
	{
		Reaction parent = timeOrderedQueue[parentIndex];
		timeOrderedQueue[childIndex] = parent;
		parent.setQueueIndex(childIndex);
		timeOrderedQueue[parentIndex] = child;
		child.setQueueIndex(parentIndex);	
	}
	
	/**
	 * Show the time-ordered queue
	 */
	private void showQueue()
	{
		int numberOfReactions = reactions.size();
		for(int index = 0; index < numberOfReactions; index++) {
		System.out.println(index + ": " + timeOrderedQueue[index].getNextTime() + " " +
				timeOrderedQueue[index]);
		}
		showQueue(0);		
	}
	
	/**
	 * Show the time-ordered queue at the given index.
	 * @param index The queue index
	 */
	private void showQueue(int index)
	{
		if(index < timeOrderedQueue.length) {
			System.out.print(timeOrderedQueue[index].getNextTime() + " " + timeOrderedQueue[index]);
			int leftChildIndex = 2 * index + 1;
			if(leftChildIndex < timeOrderedQueue.length) {
				System.out.print(timeOrderedQueue[leftChildIndex].getNextTime() + " ");
				int rightChildIndex = leftChildIndex + 1;
				if(rightChildIndex < timeOrderedQueue.length) {
					System.out.print(timeOrderedQueue[rightChildIndex].getNextTime() + " ");
					System.out.println();
					showQueue(leftChildIndex);
					showQueue(rightChildIndex);
				}
				else {
					System.out.println();
					showQueue(leftChildIndex);
				}
			}
			else {
				System.out.println();
			}
		}
		else {
			System.out.println();
		}
	}
}
