package simulation;

import java.util.ArrayList;
import java.util.List;

/**
 * Maintain a set of interacting reactions.
 * Note that this class does not record dependencies between
 * reactions, so the aZero calculations are not optimized.
 * 
 * @author David J. Barnes (d.j.barnes@kent.ac.uk)
 *
 */
public class ReactionSet
{
	// Maintain a list of the reactions.
	private List<Reaction> reactions;
	// The sum of the propensities of all reactions.
	private double aZero;
	
	/**
	 * Create a ReactionSet.
	 */
	public ReactionSet()
	{
		reactions = new ArrayList<>();
	}
		
	/**
	 * Clear the reaction set from any previous run.
	 */
	public void clear()
	{
		this.reactions.clear();
	}
	
	/**
	 * Initialise the required reaction set.
	 */
	public void init()
	{
		calculateAZero();
	}
	
	/**
	 * Sum the a-values of all the reactions.
	 * Note that this version does not attempt to limit recalculation
	 * of reaction probabilities to only those affected by the
	 * most recent reaction.
	 * @return Return the sum of the individual a-values.
	 */
	public double calculateAZero()
	{
		aZero = 0;
		for(Reaction r : reactions) {
			aZero += r.calculatePropensity();
		}
		return aZero;
	}
	
	/**
	 * Return the sum of the individual propensities.
	 * @return The sum of the individual propensities.
	 */
	public double getAZero()
	{
		return aZero;
	}
	
	/**
	 * Add a new Reaction to the set of reactions.
	 * @param The reaction to be added.
	 * @return The reaction added.
	 */
	public Reaction addReaction(Reaction r)
	{
		reactions.add(r);
		return r;
	}
	
	/**
	 * Find the reaction such that the cumulative a-values
	 * include the passed parameter. 
	 * @param aValue Reaction selector
	 * @return The selected reaction.
	 */
	public Reaction findReaction(double aValue)
	{
		double sum = 0;
		for(Reaction r : reactions) {
			sum += r.getPropensity();
			if(sum >= aValue) {
				return r;
			}
		}	
		throw new RuntimeException("Failed to find the reaction for a = " + aValue);
	}
	
}
