package simulation;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

/**
 * Maintain a group of reactions with propensity values in the
 * range (boundary/2,boundary]
 *
 * @author David J. Barnes (d.j.barnes@kent.ac.uk)
 *
 */
public class PropensityGroup 
{
	// The group's id.
	final private int id;
	// The dynamic list of reactions.
	final private List<Reaction> reactions;
	// The boundary of the previous group.
	final private double lowerBound;
	// The upper bound of this group.
	final private double boundary;
	// The sum of all propensities in this group.
	private double groupSum;
	// The group size might be smaller than the list size because reactions
	// can be removed as well as added.
	private int groupSize;

	/**
	 * Create a propensity group.
	 * @param id The group's id
	 * @param boundary The upper (inclusive) bound of propensities in this group.
	 */
	public PropensityGroup(int id, double boundary)
	{
		this.reactions = new ArrayList<Reaction>();
		this.id = id;
		if(id == 0) {
			// This is the catch-all initial group.
			this.lowerBound = 0;
		}
		else {
			this.lowerBound = boundary / 2.0;
		}
		this.boundary = boundary;
		this.groupSum = 0.0;
		this.groupSize = 0;
	}
	
	/**
	 * Add a reaction to the group.
	 * @param r
	 */
	public void addReaction(Reaction r)
	{
		// Take account of the fact that the list might not be full.
		int index = groupSize;
		if(index < reactions.size()) {
			reactions.set(index, r);
		}
		else {
			reactions.add(r);
		}
		groupSum += r.getPropensity();
		r.setGroup(this);
		r.setGroupIndex(groupSize);
		groupSize++;
	}
	
	/**
	 * Use the 'rejection' approach to select a random reaction.
	 * @return The selected reaction.
	 */
	public Reaction selectReaction(Random rand)
	{
		Reaction r;
		double propensity;
		do {
			propensity = boundary * rand.nextDouble();
			r = reactions.get(rand.nextInt(groupSize));
		} while(r.getPropensity() < propensity);
		return r;
	}
	
	/**
	 * Revise the propensity of the group according to r's new
	 * propensity.
	 * If the reaction remains in this group return true, otherwise false.
	 * @param r The reaction with revised propensity.
	 * @return true if the reaction remains in the group, otherwise false.
	 */
	public boolean revisePropensity(Reaction r)
	{
		assert reactions.contains(r);
		double oldPropensity = r.getOldPropensity();
		double propensity = r.getPropensity();
		if(propensity > lowerBound && propensity <= boundary) {
			// The reaction stays here.
			groupSum += propensity - oldPropensity;
			return true;
		}
		else {
			// Remove it from this group.
			groupSum -= oldPropensity;
			int index = r.getGroupIndex();
			// Overwrite with the last one in the group.
			if(index != groupSize - 1) {
				Reaction replacement = reactions.get(groupSize - 1);
				reactions.set(index, replacement);				
				// reactions.remove(groupSize - 1); // Not strictly necessary.
				replacement.setGroupIndex(index);
			}
			else {
				// reactions.remove(index); // Not strictly necessary.
			}
			groupSize--;
			return false;
		}
	}
	
	/**
	 * Return the cumulative sum of propensities in the group.
	 * @return the cumulative sum of propensities in the group.
	 */
	public double getGroupPropensitySum()
	{
		return groupSum;
	}
	
	/**
	 * Return the size of the group.
	 * @return
	 */
	public int size()
	{
		return groupSize;
	}
	
	/**
	 * Return a string for this group.
	 * @return Details of the group.
	 */
	public String toString()
	{
		// NB: reactions may contain removed reactions. Better to use groupSize.
		return "Group: " + id + " Propensity " + groupSum + " " + reactions;
	}
	
	/**
	 * Return the id of this group.
	 * @return The group's id.
	 */
	public int getId()
	{
		return id;
	}

	/**
	 * Internal consistency check.
	 */
	private void checkPropensity()
	{
		double sum = 0.0;
		for(int i = 0; i < groupSize; i++) {
			Reaction r = reactions.get(i);
			sum += r.getPropensity();
		}
		assert sum == groupSum || Math.abs(sum - groupSum) < 1e-5 : "Should be: " + sum + " is " + groupSum;
	}
}
